@@ -28,7 +28,9 @@ class BaseLayer:
         The next layer in the neural network. Added automatically
         during initialization.
     input_shape : tuple
-        The shape of the input array. Computed during initialization.
+        The shape of the input array. This is a required argument
+        for the first layer, otherwise computed during
+        initialization. Should have shape (N, D) or (N, D, H, W).
     weight_init : str
         The type of initialization to use when creating the layer
         weights. See initialize_weights in utils for more info.
@@ -71,16 +73,16 @@ class Dense(BaseLayer):
 
     Parameters
     ----------
-    n_outputs : int
-        Dimensionality of output space, the number of neurons.
-    n_inputs : int,default=None
-        Input dimensions. Required if first layer, otherwise
-        calculated during initialization
+    n_neurons : int
+        The number of neurons, dimensionality of output space.
+    input_shape : tuple,default=None
+        Input shape. Required if first layer, otherwise calculated
+        during initialization. Should have shape (N, D).
     """
-    def __init__(self, n_outputs, n_inputs=None, **kwargs):
+    def __init__(self, n_neurons, input_shape=None, **kwargs):
         super().__init__(**kwargs)
-        self.n_outputs = n_outputs
-        self.n_inputs = n_inputs
+        self.n_neurons = n_neurons
+        self.input_shape = input_shape
 
     def __str__(self):
         return 'Dense Layer'
@@ -90,25 +92,29 @@ class Dense(BaseLayer):
         Initialize settings to prepare the layer for training
         """
         # If layer is not first layer use prev_layer to get input_shape
-        if self.prev_layer and self.n_inputs is None:
+        if self.prev_layer:
             self.input_shape = self.prev_layer.output.shape
-            self.n_inputs = self.input_shape[-1]
-        # Initialize output array
-        self.output = np.zeros((self.n_inputs, self.n_outputs),
+        elif self.prev_layer is None and self.input_shape is None:
+            raise Exception(
+                'input_shape is a required argument for the first layer.')
+        # Initialize output array, shape: (batch-size, n_neurons)
+        self.output = np.zeros((self.input_shape[0], self.n_neurons),
                                dtype=self.dtype)
-        # Initialize weights and biases
-        self.weights = initialize_weights(self.output.shape,
+        # Initialize weights and biases.
+        # Weights have shape: (input-dimensions, output-dimensions)
+        weights_shape = (self.input_shape[-1], self.n_neurons)
+        self.weights = initialize_weights(weights_shape,
                                           self.weight_init,
                                           self.dtype)
-        self.bias = np.zeros((1, self.n_outputs),
+        self.bias = np.zeros((1, self.n_neurons),
                              dtype=self.dtype)
         # Initialize arrays to store optimizer momentum values
-        self.weight_momentum = np.zeros(self.weights.shape,
+        self.weight_momentum = np.zeros(weights_shape,
                                         dtype=self.dtype)
         self.bias_momentum = np.zeros(self.bias.shape,
                                       dtype=self.dtype)
         # Initialize arrays to store optimizer gradient cache
-        self.weight_grad_cache = np.zeros(self.weights.shape,
+        self.weight_grad_cache = np.zeros(weights_shape,
                                           dtype=self.dtype)
         self.bias_grad_cache = np.zeros(self.bias.shape,
                                         dtype=self.dtype)
@@ -373,8 +379,8 @@ class MaxPooling(BaseLayer):
 
     Downsamples input data by taking the maximum value from each
     spatial pooling window. When pooling window size and stride are
-    both 2 (default values) the resulting array is halved along height
-    and width.
+    both 2 (default values) the resulting array is halved along
+    height and width.
 
     Parameters
     ----------
@@ -398,7 +404,7 @@ class MaxPooling(BaseLayer):
         """
         Initialize settings to prepare the layer for training
         """
-        # If layer is not first layer use prev_layer to get input_shape
+        # Use prev_layer to get input_shape if not first layer
         if self.input_shape is None:
             self.input_shape = self.prev_layer.output.shape
 
@@ -422,7 +428,8 @@ class MaxPooling(BaseLayer):
 
     def get_strided_view(self, arr):
         """
-        Return a view of an array using Numpy's as_strided slide-trick.
+        Return a view of an array using Numpy's as_strided
+        slide-trick.
 
         Computationally efficient way to get all the kernel windows
         to be used in the convolution operation. Takes 4D tensor as
@@ -522,3 +529,275 @@ class Flatten(BaseLayer):
         input dimensions.
         """
         self.dinputs = grads.reshape(self.input_shape)
+
+
+class Dropout(BaseLayer):
+    """
+    Applies dropout to the input during the forward pass.
+
+    The Dropout layer creates a mask to randomly set input units
+    to zero with probabilty 'proba'. This mask is stored as an
+    attribute which is used to propogate the gradients during
+    backpropogation.
+
+    Parameters
+    ----------
+    proba: float
+        The dropout probability of neurons to keep. The dropout mask
+        is created by sampling from a Bernoulli distribution with
+        this probability.
+
+    References
+    ----------
+    - https://www.cs.toronto.edu/~rsalakhu/papers/srivastava14a.pdf
+    """
+    def __init__(self, proba, **kwargs):
+        super().__init__(**kwargs)
+        self.proba = proba
+
+    def __repr__(self):
+        return f'Dropout-{self.proba}'
+
+    def initialize(self):
+        """
+        Initialize settings to prepare the layer for training
+        """
+        self.input_shape = self.prev_layer.output.shape
+        self.output = np.zeros(self.input_shape, dtype=self.dtype)
+
+    def forward(self, X, training=True):
+        """
+        Apply dropout to the input.
+
+        Parameters
+        ----------
+        X: numpy.ndarray
+            The input tensor to which dropout is applied.
+        training : bool,default=True
+            If set to true, applies dropout to the input during
+            training mode. If set to false, no dropout is applied
+            during inference mode.
+
+        Returns
+        -------
+        output : numpy.ndarray
+            The input with dropout applied.
+        """
+        if training:
+            self.mask = np.random.binomial(1, self.proba,
+                                           size=self.input_shape) / self.proba
+            self.output = X * self.mask
+        else:
+            self.output = X.copy()
+        return self.output
+
+    def backward(self, grads):
+        """
+        Propagate gradients through the dropout mask during
+        backpropagation.
+
+        Parameters
+        ----------
+        grads: numpy.ndarray
+            The gradient of the loss with respect to the output of
+            the dropout layer.
+
+        Returns
+        -------
+        output : numpy.ndarray
+            The gradient of the loss with respect to the input of
+            the dropout layer.
+        """
+        self.dinputs = grads * self.mask
+
+
+class BatchNorm(BaseLayer):
+    """
+    Batch Normalization layer for neural networks.
+
+    Batch normalization normalizes the input values by subtracting
+    the mean and dividing by the standard deviation, so that the
+    output has zero mean and unit variance. During training, the
+    batch mean and variance are used. However, for testing and
+    inference, a running mean and variance are kept and used.
+
+    This layer can be applied to both fully-connected and
+    convolutional layers. The total number of elements are stored
+    in the 'Nt' attribute so that operations can be performed over
+    the channel dimension, regardless of input size.
+
+    Parameters
+    ----------
+    epsilon : float, default=1e-5
+        A small constant added to the variance to avoid division by
+        zero errors.
+    mom : float, default=0.9
+        The momentum for the running mean and variance.
+    kwargs:
+        Additional keyword arguments passed to the BaseLayer.
+
+    References
+    ----------
+    - https://arxiv.org/pdf/1502.03167.pdf
+    """
+    def __init__(self,  epsilon=1e-5, mom=0.9, **kwargs):
+        super().__init__(**kwargs)
+        self.epsilon = epsilon
+        self.mom = mom
+
+    def __repr__(self):
+        return 'Batch Normalization'
+
+    def initialize(self):
+        """
+        Initialize settings to prepare the layer for training.
+
+        - self.weights and self.bias represent the gamma and beta
+          parameters in the Batch Normalization paper respectively.
+        - self.dweights and self.dbiases are used to store the
+          gradients of the loss w.r.t. the gamma and beta parameters
+          respectively.
+        - self.running_mean and self.running_var store the running
+          mean and variance which are used during inference.
+        """
+        if self.input_shape is None:
+            self.input_shape = self.prev_layer.output.shape
+        # Check input-dims to set params for 2 or 4 input dims
+        if len(self.input_shape) == 4:
+            self.N, self.C, self.H, self.W = self.input_shape
+            shape = (1, self.C, 1, 1)
+            self.axes = (0, 2, 3)
+            self.keepdims = True
+            self.Nt = self.N * self.H * self.W
+        else:
+            self.N, self.C = self.input_shape
+            shape = (self.C,)
+            self.axes = 0
+            self.keepdims = False
+            self.Nt = self.N
+
+        # weights & bias represent gamma and beta in paper
+        self.weights = np.ones(shape, dtype=self.dtype)
+        self.bias = np.zeros(shape, dtype=self.dtype)
+        self.dweights = np.zeros(shape, dtype=self.dtype)
+        self.dbiases = np.zeros(shape, dtype=self.dtype)
+        # Running mean & var initialized during first forward pass
+        self.running_mean = None
+        self.running_var = None
+
+        # Initialize arrays to store optimizer momentum values
+        self.weight_momentum = np.zeros(shape, dtype=self.dtype)
+        self.bias_momentum = np.zeros(shape, dtype=self.dtype)
+        # Initialize arrays to store optimizer gradient cache
+        self.weight_grad_cache = np.zeros(shape, dtype=self.dtype)
+        self.bias_grad_cache = np.zeros(shape, dtype=self.dtype)
+
+        self.output = np.zeros(self.input_shape, dtype=self.dtype)
+
+    def forward(self, X, training=True):
+        """
+        Perform the forward pass of the batch normalization layer.
+
+        During training, the mean and variance of the current batch
+        are used to normalize the input, and the running mean and
+        variance are updated with a momentum term to keep track of
+        the historical statistics of the data. During inference, the
+        running mean and variance are used instead of the current
+        batch mean and variance to normalize the input and ensure
+        stable output across different batches.
+
+        Parameters
+        ----------
+        X : numpy.ndarray
+            The input tensor to apply batch normalization to.
+        training : bool, default=True
+            Flag indicating whether the layer is in training mode
+            or not.
+
+        Returns
+        -------
+        numpy.ndarray
+            The normalized output tensor.
+        """
+        # Running mean & var set to batch mean/var on first pass
+        if self.running_mean is None:
+            self.running_mean = np.mean(X, axis=self.axes,
+                                        keepdims=self.keepdims)
+            self.running_var = np.var(X, axis=self.axes,
+                                      keepdims=self.keepdims)
+
+        if training:
+            # If training use batch mean and batch var
+            mean = np.mean(X, axis=self.axes, keepdims=self.keepdims)
+            self.var = np.var(X, axis=self.axes, keepdims=self.keepdims)
+            # Update running mean/var
+            self.running_mean = (1 - self.mom) * mean \
+                                 + self.mom * self.running_mean
+            self.running_var = (1 - self.mom) * self.var \
+                                + self.mom * self.running_var
+        else:
+            # If test mode use running mean and var
+            mean = self.running_mean
+            self.var = self.running_var
+        # Perform forward pass and store attributes for backpass
+        self.X_mu = (X - mean)
+        self.X_norm = self.X_mu / np.sqrt(self.var + self.epsilon)
+        self.output = self.weights * self.X_norm + self.bias
+        return self.output
+
+    def backward(self, grads):
+        """
+        Perform the backward pass of the batch normalization layer.
+
+        During the backward pass, the gradient of the loss with
+        respect to the input of this layer is computed using the
+        following steps:
+
+        1. The gradients of the loss with respect to the bias and
+           weights are computed by summing the gradients along the
+           appropriate axes, as specified by the self.axes and
+           self.keepdims attributes. These gradients are stored as
+           self.dbiases and self.dweights respectively.
+        2. The total number of elements in the input, Nt, is
+           calculated based on the input shape. If the input has 4
+           dimensions (i.e. it's a convolutional layer), Nt is
+           calculated as N * H *W. If the input has 2 dimensions
+           (i.e. it's a fully-connected layer), Nt is simply N.
+        3. The gradients of the loss with respect to the inputs are
+           computed by multiplying Nt by the gradients, subtracting
+           the sum of the gradients along the appropriate axes, and
+           subtracting the product of X_mu and the sum of the
+           gradients multiplied by X_mu, all scaled by the square
+           root of the variance plus epsilon. The final result is
+           also multiplied by the weights. This is stored as
+           self.dinputs.
+
+        Parameters
+        ----------
+        grads : numpy.ndarray
+            The gradient of the loss with respect to the output of
+            this layer.
+
+        Returns
+        -------
+        numpy.ndarray
+            The gradient of the loss with respect to the input of
+            this layer.
+
+        References
+        ----------
+        - https://github.com/cthorey/CS231/blob/master/assignment2/cs231n/layers.py
+        """
+        # Get derivatives w.r.t bias and weights
+        self.dbiases = np.sum(grads, axis=self.axes, keepdims=self.keepdims)
+        self.dweights = np.sum(self.X_norm * grads, axis=self.axes,
+                               keepdims=self.keepdims)
+
+        self.dinputs = ((1.0 / self.Nt) * self.weights
+                        * np.sqrt(self.var + self.epsilon)
+                        * (self.Nt * grads - np.sum(grads, axis=self.axes,
+                                                    keepdims=self.keepdims)
+                        - self.X_mu * (self.var + self.epsilon)**-1
+                        * np.sum(grads * self.X_mu, axis=self.axes,
+                                 keepdims=self.keepdims)))
+        return self.dinputs
